#include <cassert>
#include <iomanip>
#include <iostream>
#include "ClassDecl.h"
#include "CfHeader.h"
#include "CfSource.h"
#include "CtGlobal.h"
#include "CtLocal.h"
#include "CtPointInstance.h"
#include "PrimitiveDecl.h"
#include "code.h"
#include "compiler.h"
#include "domain-decls.h"
#include "utils.h"


ClassDecl *PointUtilsDecl;

#define P(arity) (PointNDecl[arity - 1]->cType().name)
#define M(name, arity) (P(arity) + "_mt_" + (name))

/* Define Pn_construct() and Pn_empty(). */
static void createPointConstruct(ostream &hFile, int arity)
{
  hFile << "#define " << P(arity) << "_construct(p";
  for (int i = 0; i < arity; i++)
    hFile << ", v" << i;
  hFile << ") \\\n";
  hFile << "do { \\\n";
  for (int i = 0; i < arity; i++)
    hFile << "  p.x" << i << " = v" << i << "; \\\n";
  hFile << "} while (0)\n\n";

  hFile << "#define "
	<< P(arity) << "_empty() "
	<< M("all", arity) << "_jint(0)\n\n";
}

/* Point op Point. */
static void createPointOp(ostream &hFile, int arity,
			  const char *op, const char *name)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " " <<
    M(name, arity) << "_" << P(arity) << "(" << P(arity) << " p1, " <<
    P(arity) << " p2)\n{\n";
  hFile << "  " << P(arity) << " ret;\n";
  for (int i = 0; i < arity; i++)
    hFile << "  ret.x" << i << " = p1.x" << i <<
      ' ' << op << " p2.x" << i << ";\n";
  hFile << "  return ret;\n}\n";
}

/* Point op int. */
static void createPointOpI(ostream &hFile, int arity,
			   const char *op, const char *name)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " " <<
    M(name, arity) << "_jint(" << P(arity) << " p1, " <<
    "jint v)\n{\n";
  hFile << "  " << P(arity) << " ret;\n";
  for (int i = 0; i < arity; i++)
    hFile << "  ret.x" << i << " = p1.x" << i << ' ' << op << " v;\n";
  hFile << "  return ret;\n}\n";
}

static void createPointOpPrefix(ostream &hFile, int arity,
				const char *op, const char *name)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " " << 
    M(name, arity) << "_" << P(arity) << "(" << P(arity) << " p1, " <<
    P(arity) << " p2)\n{\n";
  hFile << "  " << P(arity) << " ret;\n";
  for (int i = 0; i < arity; i++)
    hFile << "  ret.x" << i << " = " << op <<
      "(p1.x" << i << ", p2.x" << i << ");\n";
  hFile << "  return ret;\n}\n";
}

/* ... is to createPointOpPrefix as createPointOpI is to createPointOp. */
static void createPointOpIPrefix(ostream &hFile, int arity,
				 const char *op, const char *name)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " " <<
    M(name, arity) << "_jint(" << P(arity) << " p1, " <<
    "jint v)\n{\n";
  hFile << "  " << P(arity) << " ret;\n";
  for (int i = 0; i < arity; i++)
    hFile << "  ret.x" << i << " = " << op << "(p1.x" << i << ", v);\n";
  hFile << "  return ret;\n}\n";
}

static void createPointComparison(ostream &hFile, int arity,
				  const char *op, const char *name)
{
  hFile << "static INLINE_MODIFIER jboolean " <<
    M(name, arity) << "_" << P(arity) << "(" << P(arity) << " p1, " <<
    P(arity) << " p2)\n{\n";
  hFile << "  return ";
  for (int i = 0; i < arity; i++) {
    if (i) hFile << " && ";
    hFile << "p1.x" << i << ' ' << op << " p2.x" << i;
  }
  hFile << ";\n}\n";
}

static void createPointIsNotEqual(ostream &hFile, int arity)
{
  hFile << "#define " << M("isNotEqual", arity) << "_" << P(arity) << "(x, y) "
    "(!" << M("isEqual", arity) << "_" << P(arity) << "((x), (y)))\n";
}

static void createPointAll(ostream &hFile, int arity)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " "
	<< M("all", arity) << "_jint(jint x)\n{\n"
	<< "  " << P(arity) << " ret;\n"
	<< "  " << P(arity) << "_construct(ret";
  for (int i = 0; i < arity; i++)
    hFile << ", x";
  hFile << ");\n  return ret;\n}\n";
  hFile << "#define " << M("setAllThis", arity) << "_jint "
    << M("all", arity) << "_jint\n";
}

static void createPointDirection(ostream &hFile, int arity)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " " << M("direction", arity)
    << "_jint_jint(jint dir, jint val)\n{\n";
  hFile << "  if (dir < 0) { val = -val; dir = -dir; }\n";
  hFile << "  return " <<
    M("set", arity) << "_jint_jint(" <<
    M("all", arity) << "_jint(0), dir-1, val);\n}\n";

  hFile << "#define " << M("direction", arity) << "_jint(dir) " <<
    M("direction", arity) << "_jint_jint((dir), 1)\n";
  hFile << "#define " << M("setDirection", arity) << "_jint " <<
    M("direction", arity) << "_jint\n";
  hFile << "#define " << M("setDirection", arity) << "_jint_jint " <<
    M("direction", arity) << "_jint_jint\n";
}

static void createPointArity(ostream &hFile, int arity)
{
  hFile << "#define " <<
    M("arity", arity) << "() (" << arity << ")\n";
}

static void createPointGet(ostream &hFile, int arity)
{
  hFile << "static INLINE_MODIFIER jint " << M("get", arity)
	<< "_jint(" << P(arity) << " p, jint dim)\n{\n"
	<< "  switch (dim) {\n";
  
  for (int dimension = 0; dimension < arity; ++dimension)
    hFile << "    case " << dimension << ":  return p.x"
	  << dimension << ";\n";
  
  hFile << "    default: \n"
	<< "      pointAssert(\"Point<" << arity << "> accessed with out of range index\", 0);\n"
	<< "      return 0;\n"
	<< "  }\n}\n"
    
	<< "#define " << M("getPrevious", arity) << "_jint(p, x) "
	<< M("get", arity) << "_jint((p), ((x) - 1))\n";
}

static void createPointSetThis(ostream &hFile, int arity)
{
  hFile << "#define " << M("setThis", arity) <<
    "_jint_jint " << M("set",arity) << "_jint_jint\n";
}

static void createPointSet(ostream &hFile, int arity)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << ' ' << M("set", arity)
	<< "_jint_jint(" << P(arity) << " p, jint dim, jint val)\n{\n"
	<< "  switch (dim) {\n";
  
  for (int dimension = 0; dimension < arity; ++dimension)
    hFile << "    case " << dimension << ":  p.x"
	  << dimension << " = val;  break;\n";
  
  hFile << "  default: \n"
	<< "    pointAssert(\"Point<" << arity << "> accessed with out of range index\", 0);\n"
	<< "  }\n"
	<< "  return p;\n"
	<< "}\n\n";
}

static void createPointMaxOrMin(ostream &hFile, int arity,
				const char *name, const char *op)
{
  hFile << "static INLINE_MODIFIER " << P(arity) << " " <<
    M(name, arity) << "_" << P(arity) << "(" << P(arity) << " p1, " <<
    P(arity) << " p2)\n{\n";
  hFile << "  " << P(arity) << " ret;\n";
  for (int i = 0; i < arity; i++)
    hFile << "  ret.x" << i << " = (p1.x" << i <<
      ' ' << op << " p2.x" << i << " ? p1.x" <<
      i << " : p2.x" << i << ");\n";
  hFile << "  return ret;\n}\n";
}

static void createPointMaxMin(ostream &hFile, int arity)
{
  createPointMaxOrMin(hFile, arity, "getUpperBound", ">");
  createPointMaxOrMin(hFile, arity, "getLowerBound", "<");
}

static void createPointPermute(ostream &hFile, int arity)
{
  /* Permute checker */
  hFile << "static INLINE_MODIFIER jboolean " <<
    M("checkPermute", arity) << "(" << P(arity) << " p)\n{\n";
  if (arity < 31) {
    /* if arity is less than 31, use a bit field to do sort */
    hFile << "  jint bitFlags = 0;\n"
	  << "  int err = 0;\n"
	  << "  int i;\n"
	  << "  for (i = 0; (i < " << arity << ") && (!err); i++) {\n"
	  << "    jint p_i = " << M("get", arity) << "_jint(p, i);\n"
	  << "    if ((p_i <= 0) || (p_i > " << arity << ")) {\n"
	  << "      err = 1;\n"
	  << "    } else {\n"
	  << "      jint mask = ((jint) 1) << (p_i-1);\n"
	  << "      if (bitFlags & mask) {\n"
	  << "        err = 1;\n"
	  << "      } else {\n"
	  << "        bitFlags = bitFlags | mask;\n"
	  << "      }\n"
	  << "    }\n"
	  << "  }\n"
	  << "  return ((bitFlags == ((((jint) 1) << " << arity << ") - ((jint) 1))) && (!err));\n";
  } else {
    /* else use an array to do O(n) sort */
    hFile << "  int flags[" << arity << "];\n"
	  << "  int count = 0;\n"
	  << "  int err = 0;\n"
	  << "  int i;\n"
	  << "  for (i = 0; i < " << arity << "; i++) flags[i] = 0;\n"
	  << "  for (i = 0; (i < " << arity << ") && (!err); i++) {\n"
	  << "    jint p_i = " << M("get", arity) << "_jint(p, i);\n"
	  << "    if ((p_i <= 0) || (p_i > " << arity << ")) {\n"
	  << "      err = 1;\n"
	  << "    } else {\n"
	  << "      if (flags[p_i-1] == 0) {\n"
	  << "        flags[p_i-1] = 1;\n"
	  << "        count++;\n"
	  << "      } else {\n"
	  << "        err = 1;\n"
	  << "      }\n"
	  << "    }\n"
	  << "  }\n"
	  << "  return ((count == " << arity << ") && (!err));\n";
  }
  hFile << "}\n";

  /* Permute routine */
  hFile << "static INLINE_MODIFIER " << P(arity) << " " <<
    M("permute", arity) << "_" << P(arity) << "(" << P(arity) << " p1, " <<
    P(arity) << " p2)\n{\n";
  hFile << "  " << P(arity) << " ret;\n";
  hFile << "#if POINTS_BOUNDS_CHECKING\n";
  hFile << "  if (!" << M("checkPermute", arity) << "(p2)) "
	<< "pointAssert(\"Point<" << arity
	<< "> does not result in a valid permute operation\", 0);\n";
  hFile << "#endif /* POINTS_BOUNDS_CHECKING */\n";
  for (int i = 0; i < arity; i++)
    hFile << "  ret.x" << i << " = " << M("getPrevious", arity) <<
      "_jint(p1, " << M("get", arity) << "_jint(p2, " << i << "));\n";
  hFile << "  return ret;\n}\n";
}

static void createUtils(ostream &hFile)
{
  hFile << "
#include <runtime-options.h>

#ifndef POINTS_BOUNDS_CHECKING
#define POINTS_BOUNDS_CHECKING 0
#endif
#if POINTS_BOUNDS_CHECKING
# include <assert.h>
# define pointAssert(context, fact) assert(((context),(fact)))
#else /* no bounds checking */
# define pointAssert(context, fact) ((void) 0)
#endif /* no bounds checking */
static INLINE_MODIFIER jint ti_gcd(jint a, jint b)
{
  return (b == 0) ? a : ti_gcd(b, a % b);
}
static INLINE_MODIFIER jint ti_lcm(jint a, jint b)
{
  return ((a == 0) || (b == 0)) ? 0 : (a * b / ti_gcd(a, b));
}
static INLINE_MODIFIER jint ti_div_round_to_minus_inf(jint a, jint b)
{
  if ((a > 0) == (b > 0)) return a / b;
  else if (b > 0) return (a - (b - 1)) / b;
  else return (a - (b + 1)) / b;
}
";
}

static void createPointToString(ostream &hFile, ostream &cFile, int arity)
{
  hFile << "extern LP_cl_java_lang_String " << M("toString", arity) << "(" << P(arity) << " var_p);\n";

  cFile << "
LP_cl_java_lang_String " << M("toString", arity) << "(" << P(arity) << " var_p)
{
  return cl_ti_domains_PointUtils_mt_toString_" << P(arity) << "(var_p);
}
";
}

void createPointMethods(CfHeader &hFile, CfSource &cFile, int arity)
{
  static bool inited = false;
  if (!inited) {
    createUtils(hFile);
    inited = true;
  }

  static const CtGlobal globalString( StringDecl->cType() );
  static const CtGlobal globalStringBuffer( StringBufferDecl->cType() );
  static const CtGlobal globalVoid( PrimitiveDecl::VoidDecl.cType() );
  static const CtLocal localString( StringDecl->cType() );
  static const CtLocal localStringBuffer( StringBufferDecl->cType() );
  static const CtLocal localVoid( PrimitiveDecl::VoidDecl.cType() );

  PointUtilsDecl->includeSelf( cFile );
  StringBufferDecl->includeSelf( cFile );
  StringBufferDecl->cType().define( cFile );
  StringDecl->cType().define( cFile );
  globalString.define( cFile );
  globalStringBuffer.define( cFile );
  globalVoid.define( cFile );
  localString.define( cFile );
  localStringBuffer.define( cFile );
  localVoid.define( cFile );
  
  (new CtPointInstance( arity ))->define( hFile );
  cFile << "#include <domains.h>\n";

  createPointConstruct(hFile, arity);
  createPointOp(hFile, arity, "+", "add");
  createPointOp(hFile, arity, "-", "sub");
  createPointOp(hFile, arity, "*", "mul");
  createPointOpPrefix(hFile, arity, "ti_div_round_to_minus_inf", "div");
  createPointOpI(hFile, arity, "*", "mul");
  createPointOpIPrefix(hFile, arity, "ti_div_round_to_minus_inf", "div");
  createPointOpPrefix(hFile, arity, "ti_lcm", "getLcm");
  createPointComparison(hFile, arity, "==", "isEqual");
  createPointComparison(hFile, arity, ">=", "isGreaterThanEqual");
  createPointComparison(hFile, arity, "<=", "isLessThanEqual");
  createPointComparison(hFile, arity, ">", "isGreaterThan");
  createPointComparison(hFile, arity, "<", "isLessThan");
  createPointIsNotEqual(hFile, arity);
  createPointAll(hFile, arity);
  createPointArity(hFile, arity);
  createPointGet(hFile, arity);
  createPointSetThis(hFile, arity);
  createPointSet(hFile, arity);
  createPointMaxMin(hFile, arity);
  createPointPermute(hFile, arity);
  createPointDirection(hFile, arity);
  createPointToString(hFile, cFile, arity);
}
