#define PAGE_GROUP_SIZE 32
#define K 4

#ifdef MEMORY_SHARED
#include <threads.h>
ti_mutex_t freepages_lock;
#endif

#ifdef REGIONPROFILE
uint32 totalbytes, maxbytes;
uint32 totalregions, maxregions;
#endif

#define ALIGN(x, n) ((x) + ((n) - 1) & ~((n) - 1))
#define RALIGNMENT 8

#include "ti-gc.h"
#include <pages.c>
#include <alloc.c>

#ifdef MEMORY_DISTRIBUTED
void *ti_dist_malloc(size_t n)
{
  void *x = ti_malloc_atomic(n);

  if (x) {
    memset(x, 0, n);
  }

  return x;
}
#endif

/* This is small enough (256k for 512megs of memory) to not be worth
   allocating dynamically (especially as unused memory pages do not "cost"
   much) */
RegionId __regionmap[MAXPAGE];

static int STATIC_DEF(rstart);

static Region STATIC_DEF(allregions);

/* Doubly linked region list management */
void addregion(Region r)
/* Effects: Adds p to the front of doubly-linked region list */
{
  r->previous = NULL;
  r->next = STATIC_REF(allregions);
  if (STATIC_REF(allregions)) STATIC_REF(allregions)->previous = r;
  STATIC_REF(allregions) = r;
}

void remregion(Region r)
/* Effects: Removes r from the doubly-linked region list */
{
  if (r->previous)
    r->previous->next = r->next;
  else
    STATIC_REF(allregions) = r->next;
  if (r->next)
    r->next->previous = r->previous;
}

void clear(void *start, uint32 size)
{
  long *clear, *clearend;

  clear = (long *)start;
  clearend = (long *)((char *)start + size);
  do *clear++ = 0;
  while (clear < clearend) ;
}

Region newregion(void)
{
  char *first = (char *)alloc_single_page(NULL);
  Region r;

#ifdef REGIONPROFILE
  totalregions++;
  if (totalregions > maxregions) maxregions = totalregions;
#endif

  /* stagger regions across cache lines a bit */
  STATIC_REF(rstart) += 64;
  if (STATIC_REF(rstart) > RPAGESIZE / K) STATIC_REF(rstart) = 0;
  r = (Region)(first + STATIC_REF(rstart) + offsetof(struct page, pagecount));
  clear(r, sizeof *r);

  return r;
}

void initregion(Region r)
{
  char *first =
    (char *)r - STATIC_REF(rstart) - offsetof(struct page, pagecount);
  int fnb = PAGENB(first);

  /* Start using page with region header as a pointer-containing page */
  r->normal.page.base = first;
  r->normal.page.allocfrom = ALIGN(STATIC_REF(rstart) + offsetof(struct page, pagecount) + sizeof(*r), RALIGNMENT);

  /* Guarantee failure for all other blocks */
  r->normal.superpage.allocfrom = K * RPAGESIZE + 1;
  r->normal.hyperpage.allocfrom = K * K * RPAGESIZE + 1;
  r->atomic.page.allocfrom = RPAGESIZE + 1;
  r->atomic.superpage.allocfrom = K * RPAGESIZE + 1;
  r->atomic.hyperpage.allocfrom = K * K * RPAGESIZE + 1;

  /* Remember that r owns this page. */
  r->normal.pages = (struct page *)first;
  __regionmap[fnb] = Region2RegionId(r);

  /* Add to list of all regions */
  addregion(r);
}

void *ralloc(Region r, size_t size)
{
  char *mem;

  if (!r) return ti_malloc(size);

#ifdef REGIONPROFILE
  totalbytes += size;
  if (totalbytes > maxbytes) maxbytes = totalbytes;
  r->bytes += size;
#endif

  size = ALIGN(size, RALIGNMENT);
  mem = (char *)qalloc(r, &r->normal, size);
  clear(mem, size);

  return mem;
}

const void *ralloc_atomic(Region r, size_t size)
{
  if (!r) return ti_malloc_atomic(size);
  else
    {
#ifdef REGIONPROFILE
      totalbytes += size;
      if (totalbytes > maxbytes) maxbytes = totalbytes;
      r->bytes += size;
#endif
      return qalloc(r, &r->atomic, ALIGN(size, RALIGNMENT));
    }
}

#if 0
static void delpage(char *deleting, char *end)
{
  while (deleting < end)
    {
      cleanup_t cln = GET_CLEANUP_FN(deleting);
      size_t s;

      if (!cln) break;
      s = cln(deleting);
      deleting += ALIGN(s, RALIGNMENT);
    }
}
#endif

static void delregion(Region r)
{
  char *p;
  unsigned long next, pnb;

#ifdef REGIONPROFILE
  totalbytes -= r->bytes;
  totalregions--;
#endif
#if 0
  /* Mark end of last page */
  if (h->normal.allocfrom < RPAGESIZE)
    *(void **)((char *)h->normal.lastpage + h->normal.allocfrom) = NULL;

  /* Scan normal pages */
  /* First page (last in list) is special (region object there) */
  for (pnb = PAGENB(h->normal.lastpage); next = nextmap[pnb]; pnb = next)
    {
      p = PAGEADDR(pnb);
      delpage(p, p + RPAGESIZE);
    }

  /* We skip region on first page */
  p = PAGEADDR(pnb);
  delpage((char *)(h + 1), p + RPAGESIZE);
#endif

#ifdef MEMORY_SHARED
  ti_mutex_lock(&freepages_lock);
#endif

  free_all_pages(r, &r->normal);
  free_all_pages(r, &r->atomic);

#ifdef MEMORY_SHARED
  ti_mutex_unlock(&freepages_lock);
#endif

  /* Remove r from region list */
  remregion(r);
}

int deleteregion(Region hp)
{
  delregion(hp);
  return 1;
}

#ifdef BACKEND_UNIPROC
/* dummy region for NULL and GC addresses. It has id 0 */
struct cl_ti_lang_Region zeroregion;
#endif

void cl_ti_lang_Region_mt_init(void)
{
  /* Dummy fn so that this file gets included. It used to contain
     what became region_init... */
}

#ifdef REGIONPROFILE
static void print_memory_usage(void)
{
  fprintf(stderr, "Current region usage %lu bytes, max %lu bytes\n",
	  totalbytes, maxbytes);
  fprintf(stderr, "Current regions %lu, max %lu\n",
	  totalregions, maxregions);
}
#endif

void region_init(void)
{
  int i;

  for (i = 0; i < MAXPAGE; i++)
    __regionmap[i] = GCREGIONID;

  STATIC_REF(rstart) = -64; /* Save 64 bytes of memory! (sometimes ;-)) */

#ifdef MEMORY_SHARED
  if (ti_mutex_init(&freepages_lock) < 0)
    {
      perror("ti_mutex_init");	
      exit(1);
    }
#endif

#ifdef BACKEND_UNIPROC
  zeroregion.id = 1; /* Shared ! */
#endif

#ifdef REGIONPROFILE
  atexit(print_memory_usage);
#endif

}

/* GC support. Mark all non-atomic pages */

void GC_push_region_roots(void)
{
  Region r;

#ifdef MEMORY_SHARED
  /* Assumes all threads are stopped */
  int i;

  for (i = 0; i < PROCS; i++)
    for (r = allregions[i]; r; r = r->next)
#else
  for (r = allregions; r; r = r->next)
#endif
    {
      struct page *p;

      for (p = r->normal.pages; p; p = p->next)
	ti_region_push_all(p, (char *)p + RPAGESIZE);
      for (p = r->normal.bigpages; p; p = p->next)
	ti_region_push_all(p, (char *)p + p->pagecount * RPAGESIZE);
    }
}
