#include <limits.h>
#include "ti-gc.h"

/* Page allocator for region-based memory management */
/* TBD: coalescing */
/* TBD: special free list for size == K */

struct page
{
  /* low order bit is 1 if previous page is allocated, 0 if not */
  struct page *next;
  /* number of pages in this allocation unit. Field does not exist if
     pagecount == 1 (ie pages list in struct allocator) */
  uint32 pagecount;
  /* Only in free pages not in the single_pages list */
  struct page *previous;
};

/* next is the value for the next field */

struct page *alloc_single_page(struct page *next);
void free_single_page(Region r, struct page *p);

struct page *alloc_pages(int n, struct page *next);
void free_pages(Region r, struct page *p);


/* a list of free individual pages */
struct page *single_pages;

/* a list of pages of size < K (only useful for single_pages) */
struct page *solitary_pages;

/* all free pages */
struct page *unused_pages;


/* Doubly linked page list management */
void addfront(struct page **list, struct page *p)
/* Effects: Adds p to the front of doubly-linked list list */
{
  p->previous = NULL;
  p->next = *list;
  if (*list) (*list)->previous = p;
  *list = p;
}

void unlink_page(struct page **list, struct page *p)
/* Effects: Remove p from its doubly linked list */
{
  if (p->previous)
    p->previous->next = p->next;
  else
    *list = p->next;
  if (p->next)
    p->next->previous = p->previous;
}


/* Multi-page allocation management */
/* -------------------------------- */

struct page *alloc_new(int n, struct page *next)
/* Assumes freepages_lock held */
{
  struct page *newp = (struct page *)ti_region_get_mem(n << RPAGELOG);

  if (!newp)
    abort();
  assert(!((long)newp & (RPAGESIZE - 1)));

  newp->next = next;
  newp->pagecount = n;
#ifndef NMEMDEBUG
  {
    int i, pnb = PAGENB(newp);

    for (i = pnb; i < pnb + n; i++)
      __regionmap[i] = FREEPAGE;
  }
#endif

  return newp;
}

struct page *alloc_split(struct page *split, int n, struct page *next)
/* Assumes freepages_lock held */
{
#ifndef NMEMDEBUG
  /* These pages had better be free */
  int i, pnb = PAGENB(split);

  assert(split->pagecount >= n);
  for (i = pnb; i < pnb + split->pagecount; i++)
    assert(__regionmap[i] == FREEPAGE);
#endif
  if (split->pagecount > n)
    {
      /* Keep first part of block */
      split->pagecount -= n;
      if (split->pagecount < K) /* move to solitary_pages */
	{
	  unlink_page(&unused_pages, split);
	  addfront(&solitary_pages, split);
	}
      /* Return latter part of block */
      split = (struct page *)((char *)split + (split->pagecount << RPAGELOG));
    }
  else
    {
      /* remove split from list */
      unlink_page(&unused_pages, split);
    }
  split->next = next;
  split->pagecount = n;

  return split;
}

struct page *alloc_pages(int n, struct page *next)
{
  struct page *best;
  int bestn;
  struct page *scan;

  assert(n >= K);

#ifdef MEMORY_SHARED
  ti_mutex_lock(&freepages_lock);
#endif

  scan = unused_pages;
  /* Find first fit */
  for (;;)
    {
      if (!scan)
	{
	  struct page *newp = alloc_new(n, next);
#ifdef MEMORY_SHARED
	  ti_mutex_unlock(&freepages_lock);
#endif
	  return newp;
	}

      if (scan->pagecount >= n) break;
      scan = scan->next;
    }

  /* Now find best fit */
  best = scan;
  bestn = scan->pagecount;
  for (;;)
    {
      scan = scan->next;
      if (!scan)
	{
	  struct page *newp = alloc_split(best, n, next);

#ifdef MEMORY_SHARED
	  ti_mutex_unlock(&freepages_lock);
#endif
	  return newp;
	}

      if (scan->pagecount >=n && scan->pagecount < bestn)
	{
	  best = scan;
	  bestn = scan->pagecount;
	}
    }
}

void free_pages(Region r, struct page *p)
/* Assumes freepages_lock held */
{
#ifndef NMEMDEBUG
  int i, pnb = PAGENB(p);
  RegionId rid = Region2RegionId(r);

  for (i = pnb; i < pnb + p->pagecount; i++)
    {
      assert(__regionmap[i] == rid);
      __regionmap[i] = FREEPAGE;
    }
#endif

  addfront(&unused_pages, p);
  /* TBD: coalesce */
}


/* Single page management */
/* ---------------------- */
/* TBD: once in a while coalesce single_pages */

void add_single_pages(struct page *base, int n)
/* Effects: Adds the group of n pages at base to the single_pages list */
{
  while (n-- > 0)
    {
      ASSERT_FREE(base);
      base->next = single_pages;
      single_pages = base;
      base = (struct page *)((char *)base + RPAGESIZE);
    }
}

void scavenge_single_pages(int n)
{
  /* Add n pages to the single_pages list */
  struct page *scan, *best;
  uint32 bestn;

  /* First, take some solitary pages */
  while (solitary_pages)
    {
      /* Don't worry if we take a little too much here */
      struct page *adding = solitary_pages;

      solitary_pages = solitary_pages->next;
      if (solitary_pages) solitary_pages->previous = NULL;
      n -= adding->pagecount;
      add_single_pages(adding, adding->pagecount);
      if (n <= 0) return;
    }
  /* We still need pages. Take any group in unused_pages that is <= n.
     Remember smallest entry > n too. This is sortof equivalent to
     a best fit where we allow partial allocations to make up a whole */
  best = NULL;
  bestn = ULONG_MAX;
  scan = unused_pages;
  while (scan)
    {
      if (scan->pagecount <= n)
	{
	  struct page *adding = scan;

	  scan = scan->next;
	  n -= adding->pagecount;
	  unlink_page(&unused_pages, adding);
	  add_single_pages(adding, adding->pagecount);
	  if (n == 0) return;
	}
      else
	{
	  if (scan->pagecount < bestn)
	    {
	      bestn = scan->pagecount;
	      best = scan;
	    }
	  scan = scan->next;
	}
    }
  /* Still not enough. Split the best block if there is one, allocate
     new pages otherwise */
  if (!best)
    add_single_pages(alloc_new(n, NULL), n);
  else
    add_single_pages(alloc_split(best, n, NULL), n);
}

struct page *alloc_single_page(struct page *next)
{
  struct page *p;

#ifdef MEMORY_SHARED
  ti_mutex_lock(&freepages_lock);
#endif

  if (!single_pages)
    {
      scavenge_single_pages(PAGE_GROUP_SIZE);
    }
  ASSERT_FREE(single_pages);
  p = single_pages;
  single_pages = p->next;
  p->next = next;

#ifdef MEMORY_SHARED
  ti_mutex_unlock(&freepages_lock);
#endif

  return p;
}

void free_single_page(Region r, struct page *p)
/* Assumes freepages_lock held */
{
#ifndef NMEMDEBUG
  ASSERT_INUSE(p, r);
  __regionmap[PAGENB(p)] = FREEPAGE;
#endif
  p->next = single_pages;
  single_pages = p;
}
